#define _GNU_SOURCE
#define _BSD_SOURCE
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sysexits.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <signal.h>

/* compile with -DUSE_FLOCK to use flock() instead of fcntl() */
#ifdef USE_FLOCK
# define TRYLOCK	LOCK_EX|LOCK_NB
# define WAITLOCK	LOCK_EX
#else
# define TRYLOCK	F_SETLK
# define WAITLOCK	F_SETLKW
#endif

static char *progname;
static void
__attribute__((format(printf,3,4)))
__attribute__((noreturn))
error(int errnum, int exitcode, const char *fmt, ...) {
  va_list ap;
  fprintf(stderr, "%s: ", progname);
  va_start(ap, fmt); vfprintf(stderr, fmt, ap); va_end(ap);
  if (errnum) fprintf(stderr, ": %s\n", strerror(errnum));
  else fputs("\n", stderr);
  exit(exitcode);
}

static const char *lckfile;
static int quiet;

static void sigalarm(int sig) {
  if (quiet)
    _exit(EX_TEMPFAIL);
  error(0, EX_TEMPFAIL,
        "lock file `%s' is already locked (timeout waiting)", lckfile);
  sig = sig;
}

int main(int argc, char **argv) {
  int fd;
  int c;
  int lck = TRYLOCK;
  int mode = O_WRONLY|O_CREAT;
  int dofork = 1;
  int waittime = 0;
#ifndef USE_FLOCK
  struct flock fl;
#endif

  if ((progname = strrchr(argv[0], '/')) == NULL) progname = argv[0];
  else argv[0] = ++progname;

  if (argc == 1) {
    printf(
"%s: execute a program with a lock set.\n"
"Usage: %s [options] lockfile program [arguments]\n"
"where options are:\n"
" -w - if the lock is already held by another process,\n"
"   wait for it to complete instead of failing immediately\n"
" -W sec - the same as -w but wait not more than sec secounds\n"
" -e - execute the program directly, no fork/wait\n"
"   (keeps extra open filedescriptor)\n"
" -n - do not create the lock file if it does not exist\n"
" -q - produce no output if lock is already held\n"
            , progname, progname);
    return 0;
  }

  while((c = getopt(argc, argv, "+wW:neq")) != EOF)
    switch(c) {
      case 'W':
        if ((waittime = atoi(optarg)) < 1)
          error(0, EX_USAGE, "invalid wait time `%s'", optarg);
      case 'w':
        lck = WAITLOCK;
        break;
      case 'n': mode &= ~O_CREAT; break;
      case 'e': dofork = 0; break;
      case 'q': quiet = 1; break;
      default: return EX_USAGE;
    }

  argc -= optind; argv += optind;
  if (argc < 2)
    error(0, EX_USAGE, "too few arguments given");

  lckfile = *argv++;

  fd = open(lckfile, mode, 0666);
  if (fd < 0)
    error(errno, EX_CANTCREAT, "unable to open `%s'", lckfile);

  if (waittime) {
    alarm(waittime);
    signal(SIGALRM, sigalarm);
  }
#ifdef USE_FLOCK
  c = flock(fd, lck);
#else
  memset(&fl, 0, sizeof(fl));
  fl.l_type = F_WRLCK;
  c = fcntl(fd, lck, &fl);
#endif
  if (waittime) alarm(0);
  if (c < 0) {
    if (errno != EWOULDBLOCK && errno != EAGAIN && errno != EACCES)
      error(errno, EX_OSERR, "unable to lock `%s'", lckfile);
    else if (quiet)
      return EX_TEMPFAIL;
    else
      error(0, EX_TEMPFAIL, "lockfile `%s' is already locked", lckfile);
  }

  if (dofork) {
    pid_t pid;
    fcntl(fd, F_SETFD, FD_CLOEXEC);
    pid = fork();
    if (pid < 0)
      error(errno, EX_OSERR, "unable to fork");
    else if (pid) {
      if (wait(&c) < 0)
        error(errno, EX_OSERR, "wait() failed");
      if (WIFSIGNALED(c))
        error(0, EX_SOFTWARE, "%s: %s", *argv, strsignal(WTERMSIG(c)));
      return WEXITSTATUS(c);
    }
  }
  execvp(*argv, argv);
  error(errno, EX_OSERR, "unable to execute %s", *argv);
}
